#!/usr/bin/env python
#
# Copyright (C) 2012 Intel Corporation
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library. If not, see <http://www.gnu.org/licenses/>.
#
# Author: Zhigang Gong <zhigang.gong@linux.intel.com>
#/

# This file is to generate inline code to lower down those builtin
# vector functions to scalar functions.
from __future__ import print_function
import re
import sys
import os

if len(sys.argv) != 4:
    print("Invalid argument {0}".format(sys.argv))
    print("use {0} spec_file_name output_file_name just_proto".format(sys.argv[0]))
    raise

all_vector = 1,2,3,4,8,16

# generate generic type sets
def gen_vector_type(type_set, vector_set = all_vector):
    ret = []
    for t in type_set:
        for i in vector_set:
            ret.append((t, i))
    return ret

def set_vector_memspace(vector_type_set, memspace):
    ret = []
    if memspace == '':
        return vector_type_set
    for t in vector_type_set:
        ret.append((t[0], t[1], memspace))
    return ret

# if we have 3 elements in the type tuple, we are a pointer with a memory space type
# at the third element.
def isPointer(t):
    return len(t) == 3

all_itype = "char","short","int","long"
all_utype = "uchar","ushort","uint","ulong"
all_int_type = all_itype + all_utype

all_float_type = "float","double","half"
all_type = all_int_type + all_float_type

# all vector/scalar types
for t in all_type:
    exec("{0}n = [\"{0}n\", gen_vector_type([\"{0}\"])]".format(t))
    exec("s{0} = [\"{0}\", gen_vector_type([\"{0}\"], [1])]".format(t))

# Predefined type sets according to the Open CL spec.
math_gentype = ["math_gentype", gen_vector_type(all_float_type)]
math_gentypef = ["math_gentypef", gen_vector_type(["float"])]
math_gentypeh = ["math_gentypeh", gen_vector_type(["half"])]
math_gentyped = ["math_gentyped", gen_vector_type(["double"])]

half_native_math_gentype = ["half_native_math_gentype", gen_vector_type(["float"])]

integer_gentype = ["integer_gentype", gen_vector_type(all_int_type)]
integer_ugentype = ["integer_ugentype", gen_vector_type(all_utype)]
integer_sgentype = ["integer_sgentype", gen_vector_type(all_int_type, [1])]

fast_integer_gentype = ["fast_integer_gentype", gen_vector_type(["uint", "int"])]

common_gentype = ["common_gentype", gen_vector_type(all_float_type)]
common_gentypef = ["common_gentypef", gen_vector_type(["float"])]
common_gentypeh = ["common_gentypeh", gen_vector_type(["half"])]
common_gentyped = ["common_gentyped", gen_vector_type(["double"])]

relational_gentype = ["relational_gentype", gen_vector_type(all_type)]
relational_igentype = ["relational_igentype", gen_vector_type(all_itype)]
relational_ugentype = ["relational_ugentype", gen_vector_type(all_utype)]

misc_gentypem = ["misc_gentypem", gen_vector_type(all_type, [2, 4, 8, 16])]
misc_gentypen = ["misc_gentypen", gen_vector_type(all_type, [2, 4, 8, 16])]
misc_ugentypem = ["misc_ugentypem", gen_vector_type(all_utype, [2, 4, 8, 16])]
misc_ugentypen = ["misc_ugentypen", gen_vector_type(all_utype, [2, 4, 8, 16])]

all_predefined_type = math_gentype, math_gentypef, math_gentyped, math_gentypeh, \
                      half_native_math_gentype, integer_gentype,integer_sgentype,\
                      integer_ugentype, charn, ucharn, shortn, ushortn, intn,    \
                      uintn, longn, ulongn, floatn, doublen, halfn, common_gentypeh, \
                      fast_integer_gentype, common_gentype, common_gentypef,     \
                      common_gentyped, relational_gentype, relational_igentype,  \
                      relational_ugentype, schar, suchar, sshort, sushort, sint, \
                      suint, slong, sulong, sfloat, shalf, sdouble, misc_gentypem,  \
                      misc_ugentypem, misc_gentypen, misc_ugentypen

# type dictionary contains all the predefined type sets.
type_dict = {}

for t in all_predefined_type:
    type_dict.update({t[0]:t[1]})

def _prefix(prefix, dtype):
    if dtype.count("gentype") != 0:
        return prefix + '_' + dtype
    return dtype

memspaces = ["__local ", "__private ", "__global ", "__generic "]

def stripMemSpace(t):
    if t[0:2] == '__':
        for memspace in memspaces :
            if t[0:len(memspace)] == memspace:
                return memspace, t[len(memspace):]
    return '', t

def check_type(types):
    for t in types:
        memspace, t = stripMemSpace(t)
        if not t in type_dict:
            print(t)
            raise TypeError("found invalid type.")

def match_unsigned(dtype):
    if dtype[0] == 'half':
        return ["ushort", dtype[1]]
    if dtype[0] == 'float':
        return ["uint", dtype[1]]
    if dtype[0] == 'double':
        return ["ulong", dtype[1]]
    if dtype[0][0] == 'u':
        return dtype
    return ['u' + dtype[0], dtype[1]]

def match_signed(dtype):
    if dtype[0] == 'half':
        return ["short", dtype[1]]
    if dtype[0] == 'float':
        return ["int", dtype[1]]
    if dtype[0] == 'double':
        return ["long", dtype[1]]
    if dtype[0][0] != 'u':
        return dtype
    return [dtype[0][1:], dtype[1]]

def match_scalar(dtype):
    return [dtype[0], 1]

# The dstType is the expected type, srcType is
# the reference type. Sometimes, the dstType and
# srcType are different. We need to fix this issue
# and return correct dst type.
def fixup_type(dstType, srcType, n):
    if dstType == srcType:
       return dstType[n]

    if dstType != srcType:
        # scalar dst type
        if len(dstType) == 1:
            return dstType[0]
        # dst is not scalar bug src is scalar
        if len(srcType) == 1:
            return dstType[n]
        if dstType == integer_sgentype[1] and srcType == integer_gentype[1]:
            return match_scalar(srcType[n])

        if dstType == integer_gentype[1] and  \
           (srcType == integer_sgentype[1] or \
            srcType == integer_ugentype[1]):
            return dstType[n]

        if dstType == integer_ugentype[1] and srcType == integer_gentype[1]:
            return match_unsigned(srcType[n])

        if dstType == relational_igentype[1] and srcType == relational_gentype[1]:
            return match_signed(srcType[n])
        if dstType == relational_ugentype[1] and srcType == relational_gentype[1]:
            return match_unsigned(srcType[n])

        if dstType == relational_gentype[1] and    \
           (srcType == relational_igentype[1] or   \
            srcType == relational_ugentype[1]):
            return dstType[n]

        if (len(dstType) == len(srcType)):
            return dstType[n]

    print(dstType, srcType)
    raise TypeError("type mispatch")

class builtinProto():
    valueTypeStr = ""
    functionName = ""
    paramTypeStrs = []
    paramCount = 0
    outputStr = []
    prefix = ""
    justproto = 0

    def init(self, sectionHeader, sectionPrefix, justproto):
        self.valueTypeStr = ""
        self.functionName = ""
        self.paramTypeStrs = []
        self.paramCount = 0
        self.justproto = justproto
        if sectionHeader != "":
            self.outputStr = [sectionHeader]
        else:
            self.outputStr = []
        if sectionPrefix != "":
            self.prefix = sectionPrefix
        self.indent = 0

    def append(self, line, nextInit = ""):
        self.outputStr.append(line);
        return nextInit;

    def indentSpace(self):
        ret = ""
        for i in range(self.indent):
            ret += ' '

        return ret

    def init_from_line(self, t):
        self.append('//{0}'.format(t))
        line = [_f for _f in re.split(',| |\(', t.rstrip(')\n')) if _f]
        self.paramCount = 0
        stripped = 0
        memSpace = ''
        for i, text in enumerate(line):
            idx = i - stripped
            if idx == 0:
                self.valueTypeStr = _prefix(self.prefix, line[i])
                continue

            if idx == 1:
                self.functionName = line[i];
                continue

            if idx % 2 == 0:
                if line[i][0] == '(':
                    tmpType = line[i][1:]
                else:
                    tmpType = line[i]
                if tmpType == '__local' or   \
                   tmpType == '__private' or \
                   tmpType == '__global' or\
                   tmpType == '__generic':
                   memSpace = tmpType + ' '
                   stripped += 1
                   continue
                self.paramTypeStrs.append(memSpace + _prefix(self.prefix, tmpType))
                memSpace = ''
                self.paramCount += 1

    def gen_proto_str_1(self, vtypeSeq, ptypeSeqs, i):
        for n in range(0, self.paramCount):
            ptype = fixup_type(ptypeSeqs[n], vtypeSeq, i);
            vtype = fixup_type(vtypeSeq, ptypeSeqs[n], i);
            # XXX FIXME now skip all double vector, as we don't
            # defined those scalar version's prototype.
            if ptype[0].find('double') != -1 or \
               vtype[0].find('double') != -1:
                return

            if (n == 0):
                formatStr = 'OVERLOADABLE {0}{1} {2} ('.format(vtype[0], vtype[1], self.functionName)
            else:
                formatStr += ', '

            if vtype[1] == 1:
                return

            if isPointer(ptype):
                formatStr += ptype[2]
                pointerStr = '*'
            else:
                pointerStr = ''

            if ptype[1] != 1:
                formatStr += '{0}{1} {2}param{3}'.format(ptype[0], ptype[1], pointerStr, n)
            else:
                formatStr += '{0} {1}param{2}'.format(ptype[0], pointerStr, n)

        formatStr += ')'
        if self.justproto == "1":
            formatStr += ';'
            self.append(formatStr)
            return formatStr
        formatStr = self.append(formatStr, '{{return ({0}{1})('.format(vtype[0], vtype[1]))
        self.indent = len(formatStr)
        for j in range(0, vtype[1]):
            if (j != 0):
                formatStr += ','
                if (j + 1) % 2 == 0:
                    formatStr += ' '
                if j % 2 == 0:
                    formatStr = self.append(formatStr, self.indentSpace())

            if self.prefix == 'relational' and self.functionName != 'bitselect' and self.functionName != 'select':
                formatStr += '-'
            formatStr += '{0}('.format(self.functionName)
            for n in range(0, self.paramCount):
                if n != 0:
                    formatStr += ', '

                ptype = fixup_type(ptypeSeqs[n], vtypeSeq, i)
                vtype = fixup_type(vtypeSeq, ptypeSeqs[n], i)
                if vtype[1] != ptype[1]:
                    if ptype[1] != 1:
                        raise TypeError("parameter is not a scalar but has different width with result value.")
                    if isPointer(ptype):
                        formatStr += '&'
                    formatStr += 'param{0}'.format(n)
                    continue

                if (isPointer(ptype)):
                    formatStr += '({0} {1} *)param{2} + {3:2d}'.format(ptype[2], ptype[0], n, j)
                else:
                    if (self.functionName == 'select' and n == 2):
                        formatStr += '({0})(param{1}.s{2:X} & (({0})1 << (sizeof({0})*8 - 1)))'.format(ptype[0], n, j)
                    else:
                        formatStr += 'param{0}.s{1:X}'.format(n, j)

            formatStr += ')'

        formatStr += '); }\n'
        self.append(formatStr)

        return formatStr

    def output(self):
        for line in self.outputStr:
            print(line)

    def output(self, outFile):
        for line in self.outputStr:
            outFile.write('{0}\n'.format(line))

    def gen_proto_str(self):
        check_type([self.valueTypeStr] + self.paramTypeStrs)
        vtypeSeq = type_dict[self.valueTypeStr]
        ptypeSeqs = []
        count = len(vtypeSeq);
        for t in self.paramTypeStrs:
            memspace,t = stripMemSpace(t)
            ptypeSeqs.append(set_vector_memspace(type_dict[t], memspace))
            count = max(count, len(type_dict[t]))

        for i in range(count):
            formatStr = self.gen_proto_str_1(vtypeSeq, ptypeSeqs, i)

        self.append("")

# save the prototypes into ocl_vector.h
specFile = open(sys.argv[1], 'r')
headerFileName = sys.argv[2]
tempHeader = open(headerFileName, 'a')
isJustProto = sys.argv[3]

tempHeader.write("//Begin from this part is autogenerated.\n")
tempHeader.write("//Don't modify it manually.\n")

functionProto = builtinProto()
for line in specFile:
    if line.isspace():
        continue
    if line[0] == '#':
        if line[1] == '#':
            sectionHeader = "//{0} builtin functions".format(line[2:].rstrip())
            sectionPrefix=(line[2:].split())[0]
        continue
    functionProto.init(sectionHeader, sectionPrefix, isJustProto)
    sectionHeader = ""
    setionPrefix = ""
    functionProto.init_from_line(line)
    functionProto.gen_proto_str()
    functionProto.output(tempHeader)

tempHeader.close()



#!/usr/bin/python
from __future__ import print_function
import os,sys,re,string

FLT_MAX_POSI='0x1.fffffep127f'
FLT_MIN_NEGA='-0x1.fffffep127f'
FLT_MIN_POSI='ldexpf(1.0, -126)'
FLT_MAX_NEGA='ldexpf(-1.0, -126)'

paraTypeList={'float':'%e','int':'%d','double':'%lf','uint':'%d','string':'%s'}

Single_Precision_ULPs={'acos' : '4' ,
                      'acospi' : '5' ,
                      'asin' : '4' ,
                      'asinpi' : '5' ,
                      'atan' : '5' ,
                      'atan2' : '6' ,
                      'atanpi' : '5' ,
                      'atan2pi' : '6' ,
                      'acosh' : '4' ,
                      'asinh' : '4' ,
                      'atanh' : '5' ,
                      'cbrt' : '2' ,
                      'ceil' : '-1' ,
                      'copysign' : '0' ,
                      'cos' : '4' ,
                      'cosh' : '4' ,
                      'cospi' : '4' ,
                      'erfc' : '16' ,
                      'erf' : '16' ,
                      'exp' : '3' ,
                      'exp2' : '3' ,
                      'exp10' : '3' ,
                      'expm1' : '3' ,
                      'fabs' : '0' ,
                      'fdim' : '-1' ,
                      'floor' : '-1' ,
                      'fma' : '-1' ,
                      'fmax' : '0' ,
                      'fmin' : '0' ,
                      'fmod' : '0' ,
                      'fract' : '-1' ,
                      'frexp' : '0' ,
                      'hypot' : '4' ,
                      'ilogb' : '0' ,
                      'ldexp' : '-1' ,
                      'log' : '3' ,
                      'log2' : '3' ,
                      'log10' : '3' ,
                      'log1p' : '2' ,
                      'logb' : '0' ,
                      'maxmag' : '0' ,
                      'minmag' : '0' ,
                      'modf' : '0' ,
                      'nan' : '0' ,
                      'nextafter' : '0' ,
                      'pow' : '16' ,
                      'pown' : '16' ,
                      'powr' : '16' ,
                      'remainder' : '0' ,
                      'remquo' : '0' ,
                      'rint' : '-1' ,
                      'rootn' : '16' ,
                      'round' : '-1' ,
                      'rsqrt' : '2' ,
                      'sin' : '4' ,
                      'sincos' : '4' ,
                      'sinh' : '4' ,
                      'sinpi' : '4' ,
                      'sqrt' : '3' ,
                      'tan' : '5' ,
                      'tanh' : '5' ,
                      'tanpi' : '6' ,
                      'tgamma' : '16' ,
                      'trunc' : '-1' ,
                      'half_cos' : '8192' ,
                      'half_divide' : '8192' ,
                      'half_exp' : '8192' ,
                      'half_exp2' : '8192' ,
                      'half_exp10' : '8192' ,
                      'half_log2' : '8192' ,
                      'half_log10' : '8192' ,
                      'half_powr' : '8192' ,
                      'half_recip' : '8192' ,
                      'half_rsqrt' : '8192' ,
                      'half_sin' : '8192' ,
                      'half_sqrt' : '8192' ,
                      'half_tan' : '8192' }

def Min_ulp(function):
    if function in Single_Precision_ULPs.keys():
         ulpValues = Single_Precision_ULPs[function]
    else:
         ulpValues = 1
    return ulpValues


def ulpUnit(ulpSize):
  return re.findall(r"([a-zA-Z_]+)",ulpSize)[0]

def ulpNum(ulpSize):
  return re.findall(r"([0-9]+)",ulpSize)[0]

def udebug(ulpSize,returnType,function):
  #ulpUnit=re.findall(r"([a-zA-Z_]+)",ulpSize)[0]
  #ulpNum=re.findall(r"([0-9]+)",ulpSize)[0]
  text='''
    static const char* INFORNAN;
    static %s ULPSIZE, ULPSIZE_FACTOR;

    float ULPSIZE_NO_FAST_MATH = %s;
    ULPSIZE_FACTOR = select_ulpsize(ULPSIZE_FAST_MATH,ULPSIZE_NO_FAST_MATH);
    bool fast_math = ULPSIZE_FACTOR == ULPSIZE_FAST_MATH;

    if (std::isinf(cpu_data[index])){
      INFORNAN="INF";
    }
    else if (std::isnan(cpu_data[index])){
      INFORNAN="NAN";
    }
    else{
       ULPSIZE=ULPSIZE_FACTOR * cl_%s((cpu_data[index] == 0) ? 1 : cpu_data[index])
               * ((ULPSIZE_FACTOR == 1) ? %s : ( (%s == 0) ? 1 : %s));
    }

#if udebug 
    if (std::isinf(cpu_data[index])){
      if (std::isinf(gpu_data[index]))
        printf("%s expect:%s\\n", log, INFORNAN);
      else
        printf_c("%s expect:%s\\n", log, INFORNAN);
      }
    else if (std::isnan(cpu_data[index])){
      if (std::isnan(gpu_data[index]))
        printf("%s expect:%s\\n", log, INFORNAN);
      else
        printf_c("%s expect:%s\\n", log, INFORNAN);
      }
    else if ((ULPSIZE >= 0 && diff <= ULPSIZE) || (ULPSIZE < 0 && diff == 0)){
      printf("%s expect:%s\\n", log, ULPSIZE);
    }
    else
      printf_c("%s expect:%s\\n", log, ULPSIZE);
#else
    if (std::isinf(cpu_data[index])){
      sprintf(log, "%s expect:%s\\n", log, INFORNAN);
      OCL_ASSERTM(std::isinf(gpu_data[index]) || fast_math,log);
    }
    else if (std::isnan(cpu_data[index])){
      sprintf(log, "%s expect:%s\\n", log, INFORNAN);
      OCL_ASSERTM(std::isnan(gpu_data[index]) || fast_math,log);
    }
    else{
      sprintf(log, "%s expect:%s\\n", log, ULPSIZE);
      if (ULPSIZE < 0)
            OCL_ASSERTM(gpu_data[index] == cpu_data[index], log);
      else
            OCL_ASSERTM(fabs(gpu_data[index]-cpu_data[index]) <= ULPSIZE, log);
    }
#endif
  }
}\n'''%(returnType,Min_ulp(function),\
        ulpUnit(ulpSize),ulpNum(ulpSize),\
        ulpNum(ulpSize), ulpNum(ulpSize),\
        paraTypeList['string'],paraTypeList['string'],\
        paraTypeList['string'],paraTypeList['string'],\
        paraTypeList['string'],paraTypeList['string'],\
        paraTypeList['string'],paraTypeList['string'],\
        paraTypeList['string'],paraTypeList['%s'%(returnType)],\
        paraTypeList['string'],paraTypeList['%s'%(returnType)],\
        paraTypeList['string'],paraTypeList['string'],\
        paraTypeList['string'],paraTypeList['string'],\
        paraTypeList['string'],paraTypeList['%s'%(returnType)])

  return text

def gene2ValuesLoop(values1,values2,inputValues):
  values2=values2+inputValues*len(inputValues)

  for i in inputValues:
    for j in range(0,len(inputValues)):
      values1 += [i]

  return values1,values2

def gene3ValuesLoop(values1,values2,values3,inputValues):
  for i in inputValues:
    for j in range(0,len(inputValues)):
      for k in range(0,len(inputValues)):
        values1 += [i]

  for i in inputValues:
    for j in inputValues:
      for k in range(0,len(inputValues)):
        values2 += [j]

  values3=inputValues*(len(inputValues)**2)
  return values1,values2,values3

class func:
  """ This class will define all needed instance attribute in fundation a c programing file. """

  def __init__(self,name,cpuFuncName,inputType,outputType,values,ulp, cpu_func=''):
    self.funcName = name
    self.cpuFuncName = cpuFuncName
    self.fileName = 'builtin_'+name
    self.inputtype = inputType
    self.outputtype = outputType
    self.values = values
    self.ulp = ulp
    self.cpufunc=cpu_func
    self.cpplines = []
    
#####cpp file required information:
    self.Head='''/*
This file is generated by utest_generator.py.
Usually you need NOT modify this file manually.
But when any bug occured, you can change the value of udebug from 0 to 1,
which can print more values and information to assist debuging the issue.
*/

#include "utest_helper.hpp"
#include <stdio.h>
#include <cmath>
#include <algorithm>
#include <string.h>

#define udebug 0
#define FLT_MAX 0x1.fffffep127f
#define FLT_MIN 0x1.0p-126f
#define INT_ULP 0

#define printf_c(...) \\
{\\
  printf("\\033[1m\\033[40;31m");\\
  printf( __VA_ARGS__ );\\
  printf("\\033[0m");\\
}
'''
    #########Execute class itself
    self.geneProcess()

#####Computer vector && argument type:
  def argtype(self,paraN,index):
    return re.findall(r"[a-zA-Z_]+",self.inputtype[paraN][index])[0]

  def argvector(self,paraN,index):
    vector=re.findall(r"[0-9]+",self.inputtype[paraN][index])
    if vector:
      vector=int(vector[0])
    else:
      vector=1
    return vector

  def returnVector(self,index):
    returnVector=re.findall(r"[0-9]+",self.outputtype[index])
    if returnVector:
      returnVector=returnVector[0]
    else:
      returnVector=1
    return returnVector

  def retType(self,index):
    return re.findall("[a-zA-Z_]+",self.outputtype[index])[0]

  def inputNumFormat(self,paraN,index):
    return paraTypeList['%s'%(self.argtype(paraN,index))]

  def outputNumFormat(self,index):
    return paraTypeList['%s'%(self.retType(index))]

#####Cpu values analyse
  def GenInputValues(self,index):
    #namesuffix=self.inputtype[0][index]
    vlen = self.argvector(self.inputtype.__len__()-1,index)
    for i in range(0,self.values.__len__()):
      vals = []
      for j in range(0, vlen):
        if (len(vals) >= 128):	#avoid too many data
          vals = vals[0:128]
          break
        vals += self.values[i]
      self.cpplines += [ "%s input_data%d[] = {%s};" %(self.argtype(i,index),i+1,str(vals).strip('[]').replace('\'','')) ]
    self.cpplines += [ "const int count_input = sizeof(input_data1) / sizeof(input_data1[0]);" ]
    self.cpplines += [ "int vector = %s;\n"%(vlen) ]

#####Cpu Function
  def GenCpuCompilerMath(self,index):
    #namesuffix=self.inputtype[0][index]
    defline='static void cpu_compiler_math(%s *dst, '%(self.retType(index))
    cpufunargs='('
    funcline = ['{']
    vectorargs=[]

    if (self.returnVector(index) == 1 and self.argvector(0,index) != 1):
      for i in range(0,self.values.__len__()):
        defline += 'const %s *src%d'%(self.argtype(i,index),i+1)
        defline += ( i == self.values.__len__()-1 ) and ')' or ','
        vectorargs.append('(')
      for i in range(0,self.values.__len__()):
        for j in range(0,self.vector):
          vectorargs += "x%d%d"%(i+1,j+1)
          vectorargs += ( j == self.vector-1 ) and ');' or ','
          funcline += ["  const %s x%d%d = *(src%d+%d);"%(self.argtype(i,index),i+1,j+1,i+1,j)]

      return 0

    for i in range(0,self.values.__len__()):
      defline += 'const %s *src%d'%(self.argtype(i,index),i+1)
      defline += ( i == self.values.__len__()-1 ) and ')' or ','
      cpufunargs += "x%d"%(i+1)
      cpufunargs += ( i == self.values.__len__()-1 ) and ');' or ','
      funcline += ["  const %s x%d = *src%d;"%(self.argtype(i,index),i+1,i+1)]

    funcline += [ "  dst[0] = %s%s"%(self.cpuFuncName, cpufunargs) ]
    funcline += [ '}'] 

    funcline = [defline] + funcline

    self.cpplines += funcline
#    self.writeCPP( '\n'.join(funcline), 'a', namesuffix)

  def writeCPP(self,content,authority,namesuffix):
    file_object = open("generated/%s_%s.cpp"%(self.fileName,namesuffix),authority)
    file_object.writelines(content)
    file_object.close()

  def writeCL(self,content,authority,namesuffix):
    file_object = open(os.getcwd()+"/../kernels/%s_%s.cl"%(self.fileName,namesuffix),authority)
    file_object.writelines(content)
    file_object.close()

  def nameForCmake(self,content,namesuffix):
    print("generated/%s_%s.cpp"%(self.fileName,namesuffix),end=" ")

  def utestFunc(self,index):
    funcLines=[]
    namesuffix=self.inputtype[0][index]
    funcline=[]
    funchead='''
static void %s_%s(void)
{
  int index;
  %s gpu_data[count_input] = {0}, cpu_data[count_input] = {0}, diff=0.0;
  char log[1024] = {0};

  OCL_CREATE_KERNEL(\"%s_%s\");
  OCL_CREATE_BUFFER(buf[0], CL_MEM_READ_WRITE, count_input * sizeof(%s), NULL); 

  globals[0] = count_input / vector;
  locals[0] = 1;
 '''%(self.fileName,namesuffix,\
     self.retType(index),\
     self.fileName, namesuffix,\
     self.retType(index))

    funcline += [funchead]
    for i in range(1,self.values.__len__()+1): 
      funcline += ["  OCL_CREATE_BUFFER(buf[%d], CL_MEM_READ_WRITE, count_input * sizeof(%s), NULL);"%(i,self.argtype(i-1,index))]
      funcline += ["  clEnqueueWriteBuffer( queue, buf[%d], CL_TRUE, 0, count_input * sizeof(%s), input_data%d, 0, NULL, NULL);"%(i,self.argtype(i-1,index),i)]

    funcline += ["  OCL_CREATE_BUFFER(buf[%d], CL_MEM_READ_WRITE, sizeof(int), NULL);"%(self.inputtype.__len__()+1)]
    funcline += ["  clEnqueueWriteBuffer( queue, buf[%d], CL_TRUE, 0, sizeof(int), &vector, 0, NULL, NULL);"%(self.inputtype.__len__()+1)]

	#0=output 1=input1 2=input2 ... len+2=output
    for i in range(0,self.values.__len__()+2): 
      funcline += ["  OCL_SET_ARG(%d, sizeof(cl_mem), &buf[%d]);"%(i,i)]

    funcrun='''
  // Run the kernel:
  //int errRead = clEnqueueReadBuffer( queue, buf[0], CL_TRUE, 0, sizeof(%s) * count_input, gpu_data, 0, NULL, NULL);
  OCL_NDRANGE( 1 );
  OCL_MAP_BUFFER(0);
'''%(self.argtype(0,index))
    funcline += [ funcrun ]

    text = ''' memcpy(gpu_data, buf_data[0], sizeof(gpu_data)); '''
    funcline += [ text ]

    funcsprintfa='    sprintf(log, \"'
    funcsprintfb=''
    if (self.returnVector(index) == 1 and self.argvector(0,index) != 1):
      funccompare='''
  for (index = 0; index < count_input/vector; index++)
  {
    cpu_compiler_math( cpu_data + index, '''
    else:
      funccompare='''
  for (index = 0; index < count_input; index++)
  {
    cpu_compiler_math( cpu_data + index,'''

    for i in range(0,self.values.__len__()):
      funccompare += " input_data%d + index"%(i+1)
      funccompare += (self.values.__len__() - 1 == i) and ');' or ','

      funcsprintfa += "input_data%d:"%(i+1)
      funcsprintfa += "%s "%(self.inputNumFormat(i,index))
      funcsprintfb += " input_data%d[index],"%(i+1)

    funcline += [ funccompare ]

    funcsprintfa += " -> gpu:%s  cpu:%s diff:%s\","%(self.outputNumFormat(index),self.outputNumFormat(index),self.outputNumFormat(index))#,self.outputNumFormat(index))
    funcsprintfb += " gpu_data[index], cpu_data[index], diff);"#%(ulpUnit(self.ulp),ulpNum(self.ulp))

    #funcdiff = "    diff = fabs((gpu_data[index]-cpu_data[index])"
    #funcdiff += (self.retType(index) == "int") and ');' or '/(cpu_data[index]>1?cpu_data[index]:1));'
    valuejudge = "    if (std::fpclassify(gpu_data[index]) == FP_SUBNORMAL){ gpu_data[index] = 0; }\n"
    valuejudge += "    if (std::fpclassify(cpu_data[index]) == FP_SUBNORMAL){ cpu_data[index] = 0; }\n"
    funcdiff = "    diff = fabs((gpu_data[index]-cpu_data[index]));"
    funcline += [ valuejudge ]
    funcline += [ funcdiff ]
    funcline += [ funcsprintfa + funcsprintfb ]

    self.cpplines += funcline

    self.cpplines += [ udebug(self.ulp,self.retType(index),self.funcName) ]
    self.cpplines += [ "MAKE_UTEST_FROM_FUNCTION(%s_%s)"%(self.fileName,namesuffix) ]

  def genCL(self,index):
    namesuffix=self.inputtype[0][index]
    clLine = []
    clhead = '__kernel void %s_%s(__global %s *dst, '%(self.fileName,namesuffix,self.retType(index))
    clvalueDef=''
    clcomputer=''
    tmp=''

    for i in range(0,self.values.__len__()):
      clhead += ' __global %s *src%d,'%(self.argtype(i,index),i+1)
      clvalueDef +=   '  %s x%d = (%s) ('%(self.inputtype[i][index],i+1,self.inputtype[i][index])
      tmp = 'src%d[i * (*vector) + '%(i+1)
      for j in range(0,self.argvector(i,index)):
        clvalueDef += tmp + ((self.argvector(i-1,index) == j+1 ) and '%d]);\n'%(j) or '%d],'%(j))
      clcomputer += (self.values.__len__() == i+1) and 'x%d);'%(i+1) or 'x%d,'%(i+1)
      
    clhead += ' __global int *vector) {\n'
    clhead += '  int i = get_global_id(0);'
    clLine += [ clhead ]
    clLine += [ clvalueDef ]
    clLine += [ '  %s ret;'%(self.outputtype[index]) ]
    clLine += [ '  ret = %s('%(self.funcName) + clcomputer ] 

    if (int(self.returnVector(index)) == 1):
      clLine += [ '  dst[i] = ret;' ]
    else:
      for i in range(0,int(self.returnVector(index))):
        clLine += [ '  dst[i * (*vector) + %d] = ret[%d];'%(i,i) ]
    clLine += [ '};' ]

    self.writeCL('\n'.join(clLine),'w',namesuffix)
  
  def geneProcess(self):
    for i in range(0,self.inputtype[0].__len__()):
##########Write Cpp file          
      namesuffix=self.inputtype[0][i]
      self.cpplines = []
      #The head:
      self.cpplines += [self.Head]

      self.cpplines += ["namespace {\n"]

      #Parameters:
      self.GenInputValues(i)

      #cpu function generator:
      self.cpplines += [self.cpufunc]

      #Cpu function:
      self.GenCpuCompilerMath(i)

      #utest function
      self.utestFunc(i)

      self.cpplines += ["}\n"]

      #kernel cl
      self.genCL(i)

      #CMakelists.txt
      self.nameForCmake(self.fileName,namesuffix)

      self.writeCPP( '\n'.join(self.cpplines) ,'w',namesuffix)
#########End

#def main():
#
#if __name__ == "__main__":
#  main()



#!/usr/bin/python
from utest_generator import *
import os,sys

#base_input_values = [80, -80, 3.14, -3.14, -0.5, 0.5, 1, -1, 0.0,6,-6,1500.24,-1500.24]
#extend_input_values = [FLT_MAX_POSI,FLT_MIN_NEGA,FLT_MIN_POSI,FLT_MAX_NEGA,80, -80, 3.14, -3.14, -0.5, 0.5, 1, -1, 0.0,6,-6,1500.24,-1500.24]

#func:
#    gpufuncName 
#    cpuFuncName
#    fileName: 'builtin_'+name
#    inputtype: a 2-D list because there're more than one input data
#    outputtype: a list
#    values
#    ulp

# reduce pi*x limitation to [-pi,pi]
reduce1='''
static float reduce1( float x )
{
  SF fx, fy;
  fx.f = fy.f = x;
  int n;

  fy.spliter.exponent = fx.spliter.exponent - 1;
  n = (int)fy.f;

  fx.f = fx.f - 2.0 * n;

  // reduce to [-1.0, 1.0]
  fx.f = (fx.f < -1)?(fx.f + 2.0):((fx.f > 1)?(fx.f - 2.0):fx.f);

  return fx.f;
}
'''
# define fuction: cospi
cospi='''
static float cospi(float x){
  float r = x;
  if ( x > 1 || x < -1) r = reduce1(x);

  // reduce to [0.0, 1.0]
  if (r < 0)
    r = fabs(r);

  if (r >= 0 && r <= 0.25)
    return  cosf(r * M_PI);
  else if (r > 0.25 && r <= 0.5)
    return  sinf((0.5 - r) * M_PI);
  else if (r > 0.5 && r <= 0.75)
    return sinf(-(r-0.5) * M_PI);
  else if (r > 0.75 && r <= 1.0){
    return -cosf((1 -  r) * M_PI);}

  // Error return
  return 0xffffffff;
}
'''
# define function: sinpi
sinpi='''
static float sinpi(float x){
  float r = x;
  if ( x > 1 || x < -1) r = reduce1(x);

  // reduce to [-0.5, 0.5]
  if (r < -0.5)
    r = -1 - r;
  else if (r > 0.5)
    r = 1 - r;

  if (r > 0.25 && r <= 0.5)
    return  cosf((0.5 - r) * M_PI);
  else if (r >= 0 && r <= 0.25)
    return  sinf(r * M_PI);
  else if (r >= -0.25 && r < 0)
    return -sinf(r * -M_PI);
  else if (r >= -0.5 && r < -0.25){
    return -cosf((0.5 + r) * M_PI);}

  // Error return
  return 0xffffffff;
}
'''

base_input_values = [ 0, 1, 3.14, 5.15, 6.01, 7.89]
base_input_values1 = [ 1, 3.14, 5.15, 6.01, 7.89]
def main():
  ##### gentype acos(gentype)
  acos_input_values = base_input_values
  acos_input_type = ['float','float2','float4','float8','float16']
  acos_output_type = ['float','float2','float4','float8','float16']
  acosUtests = func('acos','acos',[acos_input_type],acos_output_type,[acos_input_values],'4 * FLT_ULP')
  
  ##### gentype acosh(gentype)
  acosh_input_values = base_input_values
  acosh_input_type = ['float','float2','float4','float8','float16']
  acosh_output_type = ['float','float2','float4','float8','float16']
  acoshUtests = func('acosh','acosh',[acosh_input_type],acosh_output_type,[acosh_input_values],'4 * FLT_ULP')
  
  ##### gentype acospi(gentype x)
  acospi_input_values = base_input_values
  acospi_input_type = ['float','float2','float4','float8','float16']
  acospi_output_type = ['float','float2','float4','float8','float16']
  acospi_cpu_func='''
static float acospi(float x){
  return acos(x)/M_PI;
} '''
  acospiUtests = func('acospi','acospi',[acospi_input_type],acospi_output_type,[acospi_input_values],'4 * FLT_ULP',acospi_cpu_func)
  
  ##### gentype asin(gentype)
  asin_input_values = base_input_values
  asin_input_type = ['float','float2','float4','float8','float16']
  asin_output_type = ['float','float2','float4','float8','float16']
  asinUtests = func('asin','asin',[asin_input_type],asin_output_type,[asin_input_values],'4 * FLT_ULP')
  
  ##### gentype asinh(gentype)
  asinh_input_values = base_input_values
  asinh_input_type = ['float','float2','float4','float8','float16']
  asinh_output_type = ['float','float2','float4','float8','float16']
  asinhUtests = func('asinh','asinh',[asinh_input_type],asinh_output_type,[asinh_input_values],'4 * FLT_ULP')
  
  ##### gentype asinpi(gentype x)
  asinpi_input_values = base_input_values
  asinpi_input_type = ['float','float2','float4','float8','float16']
  asinpi_output_type = ['float','float2','float4','float8','float16']
  asinpi_cpu_func='''
static float asinpi(float x){
  return asin(x)/M_PI;
} '''
  asinpiUtests = func('asinpi','asinpi',[asinpi_input_type],asinpi_output_type,[asinpi_input_values],'4 * FLT_ULP',asinpi_cpu_func)
  
  ##### gentype atan(gentype y_over_x)
  atan_input_values = base_input_values
  atan_input_type = ['float','float2','float4','float8','float16']
  atan_output_type = ['float','float2','float4','float8','float16']
  atanUtests = func('atan','atan',[atan_input_type],atan_output_type,[atan_input_values],'5 * FLT_ULP')
  
  ##### gentype atan2(gentype y, gentype x)
  atan2_base_values = base_input_values
  atan2_input_values1 = []
  atan2_input_values2 = []
  atan2_input_values1,atan2_input_values2=gene2ValuesLoop(atan2_input_values1,atan2_input_values2,atan2_base_values)
  atan2_input_type1 = ['float','float2','float4','float8','float16']
  atan2_input_type2 = ['float','float2','float4','float8','float16']
  atan2_output_type = ['float','float2','float4','float8','float16']
  atan2Utests = func('atan2','atan2',[atan2_input_type1,atan2_input_type2],atan2_output_type,[atan2_input_values1,atan2_input_values2],'6 * FLT_ULP')
  
  ##### gentype atanh(gentype)
  atanh_input_values = base_input_values
  atanh_input_type = ['float','float2','float4','float8','float16']
  atanh_output_type = ['float','float2','float4','float8','float16']
  atanhUtests = func('atanh','atanh',[atanh_input_type],atanh_output_type,[atanh_input_values],'5 * FLT_ULP')
  
  ##### gentype atanpi(gentype x)
  atanpi_input_values = base_input_values
  atanpi_input_type = ['float','float2','float4','float8','float16']
  atanpi_output_type = ['float','float2','float4','float8','float16']
  atanpi_cpu_func='''
static float atanpi(float x){
  return atan(x)/M_PI;
} '''
  atanpiUtests = func('atanpi','atanpi',[atanpi_input_type],atanpi_output_type,[atanpi_input_values],'4 * FLT_ULP',atanpi_cpu_func)
  
  ##### gentype atan2pi(gentype y, gentype x)
  atan2pi_base_values = base_input_values
  atan2pi_input_values1 = []
  atan2pi_input_values2 = []
  atan2pi_input_values1,atan2pi_input_values2=gene2ValuesLoop(atan2pi_input_values1,atan2pi_input_values2,atan2pi_base_values)
  atan2pi_input_type1 = ['float','float2','float4','float8','float16']
  atan2pi_input_type2 = ['float','float2','float4','float8','float16']
  atan2pi_output_type = ['float','float2','float4','float8','float16']
  atan2pi_cpu_func='''
static float atan2pi(float y, float x){
  return atan2(y,x)/M_PI;
} '''
  atan2piUtests = func('atan2pi','atan2pi',[atan2pi_input_type1,atan2pi_input_type2],atan2pi_output_type,[atan2pi_input_values1,atan2pi_input_values2],'6 * FLT_ULP',atan2pi_cpu_func)
  
  ##### gentype cbrt(gentype)
  cbrt_input_values = base_input_values
  cbrt_input_type = ['float','float2','float4','float8','float16']
  cbrt_output_type = ['float','float2','float4','float8','float16']
  cbrtUtests = func('cbrt','cbrt',[cbrt_input_type],cbrt_output_type,[cbrt_input_values],'4 * FLT_ULP')
  
  ##### gentype ceil(gentype)
  ceil_input_values = base_input_values
  ceil_input_type = ['float','float2','float4','float8','float16']
  ceil_output_type = ['float','float2','float4','float8','float16']
  ceilUtests = func('ceil','ceil',[ceil_input_type],ceil_output_type,[ceil_input_values],'0 * FLT_ULP')
  
  ##### gentype copysign(gentype x, gentype y)
  copysign_base_values = base_input_values
  copysign_input_values1 = []
  copysign_input_values2 = []
  copysign_input_values1,copysign_input_values2=gene2ValuesLoop(copysign_input_values1,copysign_input_values2,copysign_base_values)
  copysign_input_type1 = ['float','float2','float4','float8','float16']
  copysign_input_type2 = ['float','float2','float4','float8','float16']
  copysign_output_type = ['float','float2','float4','float8','float16']
  copysignUtests = func('copysign','copysign',[copysign_input_type1,copysign_input_type2],copysign_output_type,[copysign_input_values1,copysign_input_values2],'0 * FLT_ULP')
  
  ##### gentype cos(gentype)
  cos_input_values = base_input_values
  cos_input_type = ['float','float2','float4','float8','float16']
  cos_output_type = ['float','float2','float4','float8','float16']
  cosUtests = func('cos','cos',[cos_input_type],cos_output_type,[cos_input_values],'4 * FLT_ULP')
  
  ##### gentype cosh(gentype)
  cosh_input_values = base_input_values
  cosh_input_type = ['float','float2','float4','float8','float16']
  cosh_output_type = ['float','float2','float4','float8','float16']
  coshUtests = func('cosh','cosh',[cosh_input_type],cosh_output_type,[cosh_input_values],'4 * FLT_ULP')
  
  ##### gentype cospi(gentype x)
  cospi_input_values = base_input_values
  cospi_input_type = ['float','float2','float4','float8','float16']
  cospi_output_type = ['float','float2','float4','float8','float16']
  cospi_cpu_func=reduce1+cospi
  cospiUtests = func('cospi','cospi',[cospi_input_type],cospi_output_type,[cospi_input_values],'2 * FLT_ULP',cospi_cpu_func)
  
  ##### gentype erf(gentype)
  erf_input_values = base_input_values
  erf_input_type = ['float','float2','float4','float8','float16']
  erf_output_type = ['float','float2','float4','float8','float16']
  erfUtests = func('erf','erf',[erf_input_type],erf_output_type,[erf_input_values],'16 * FLT_ULP')

  ##### gentype erfc(gentype)
  erfc_input_values = base_input_values
  erfc_input_type = ['float','float2','float4','float8','float16']
  erfc_output_type = ['float','float2','float4','float8','float16']
  erfcUtests = func('erfc','erfc',[erfc_input_type],erfc_output_type,[erfc_input_values],'16 * FLT_ULP')
  
  ##### gentype exp(gentype x)
  exp_input_values = base_input_values
  exp_input_type = ['float','float2','float4','float8','float16']
  exp_output_type = ['float','float2','float4','float8','float16']
  expUtests = func('exp','exp',[exp_input_type],exp_output_type,[exp_input_values],'4 * FLT_ULP')
  
  ##### gentype exp2(gentype)
  exp2_input_values = base_input_values
  exp2_input_type = ['float','float2','float4','float8','float16']
  exp2_output_type = ['float','float2','float4','float8','float16']
  exp2Utests = func('exp2','exp2',[exp2_input_type],exp2_output_type,[exp2_input_values],'4 * FLT_ULP')
  
  ##### gentype exp10(gentype)
  exp10_input_values = base_input_values
  exp10_input_type = ['float','float2','float4','float8','float16']
  exp10_output_type = ['float','float2','float4','float8','float16']
  exp10Utests = func('exp10','exp10',[exp10_input_type],exp10_output_type,[exp10_input_values],'4 * FLT_ULP')
  
  ##### gentype expm1(gentype x)
  expm1_input_values = base_input_values
  expm1_input_type = ['float','float2','float4','float8','float16']
  expm1_output_type = ['float','float2','float4','float8','float16']
  expm1Utests = func('expm1','expm1',[expm1_input_type],expm1_output_type,[expm1_input_values],'4 * FLT_ULP')
  
  ##### gentype fabs(gentype)
  fabs_input_values = base_input_values
  fabs_input_type = ['float','float2','float4','float8','float16']
  fabs_output_type = ['float','float2','float4','float8','float16']
  fabsUtests = func('fabs','fabs',[fabs_input_type],fabs_output_type,[fabs_input_values],'0 * FLT_ULP')
  
  ##### gentype fdim(gentype x, gentype y)
  fdim_base_values = base_input_values
  fdim_input_values1 = []
  fdim_input_values2 = []
  fdim_input_values1,fdim_input_values2=gene2ValuesLoop(fdim_input_values1,fdim_input_values2,fdim_base_values)
  fdim_input_type1 = ['float','float2','float4','float8','float16']
  fdim_input_type2 = ['float','float2','float4','float8','float16']
  fdim_output_type = ['float','float2','float4','float8','float16']
  fdimUtests = func('fdim','fdim',[fdim_input_type1,fdim_input_type2],fdim_output_type,[fdim_input_values1,fdim_input_values2],'0 * FLT_ULP')
  
  ##### gentype floor(gentype)
  floor_input_values = base_input_values
  floor_input_type = ['float','float2','float4','float8','float16']
  floor_output_type = ['float','float2','float4','float8','float16']
  floorUtests = func('floor','floor',[floor_input_type],floor_output_type,[floor_input_values],'0 * FLT_ULP')
  
  ##### gentype fmax(gentype x, gentype y)
  fmax_base_values = base_input_values
  fmax_input_values1 = []
  fmax_input_values2 = []
  fmax_input_values1,fmax_input_values2=gene2ValuesLoop(fmax_input_values1,fmax_input_values2,fmax_base_values)
  fmax_input_type1 = ['float','float2','float4','float8','float16']
  fmax_input_type2 = ['float','float2','float4','float8','float16']
  fmax_output_type = ['float','float2','float4','float8','float16']
  fmaxUtests = func('fmax','fmax',[fmax_input_type1,fmax_input_type2],fmax_output_type,[fmax_input_values1,fmax_input_values2],'0 * FLT_ULP')
  
  ##### gentypef fmax(gentypef x, float y)
#  fmax_gentypef_base_values = base_input_values
#  fmax_gentypef_input_values1 = []
#  fmax_gentypef_input_values2 = []
#  fmax_gentypef_input_values2,fmax_gentypef_input_values1=gene2ValuesLoop(fmax_gentypef_input_values1,fmax_gentypef_input_values2,fmax_gentypef_base_values)
#  fmax_gentypef_input_type1 = ['float','float2','float4','float8','float16']
#  fmax_gentypef_input_type2 = ['float','float','float','float','float']
#  fmax_gentypef_output_type = ['float','float2','float4','float8','float16']
#  ##### gentypef fmax(gentypef x, float y)
#  fmax_gentypefUtests = func('gentypef_fmax','gentypef_fmax',[fmax_gentypef_input_type1,fmax_gentypef_input_type2],fmax_gentypef_output_type,[fmax_gentypef_input_values1,fmax_gentypef_input_values2],'0 * FLT_ULP')
  
  ##### gentype fmin(gentype x, gentype y)
  fmin_base_values = base_input_values
  fmin_input_values1 = []
  fmin_input_values2 = []
  fmin_input_values1,fmin_input_values2=gene2ValuesLoop(fmin_input_values1,fmin_input_values2,fmin_base_values)
  fmin_input_type1 = ['float','float2','float4','float8','float16']
  fmin_input_type2 = ['float','float2','float4','float8','float16']
  fmin_output_type = ['float','float2','float4','float8','float16']
  fminUtests = func('fmin','fmin',[fmin_input_type1,fmin_input_type2],fmin_output_type,[fmin_input_values1,fmin_input_values2],'0 * FLT_ULP')
  
#  ##### gentypef fmin(gentypef x, float y)
#  fmin_gentypef_base_values = base_input_values
#  fmin_gentypef_input_values1 = []
#  fmin_gentypef_input_values2 = []
#  fmin_gentypef_input_values2,fmin_gentypef_input_values1=gene2ValuesLoop(fmin_gentypef_input_values1,fmin_gentypef_input_values2,fmin_gentypef_base_values)
#  fmin_gentypef_input_type1 = ['float','float2','float4','float8','float16']
#  fmin_gentypef_input_type2 = ['float','float','float','float','float']
#  fmin_gentypef_output_type = ['float','float2','float4','float8','float16']
#  ##### gentypef fmin(gentypef x, float y)
#  fmin_gentypefUtests = func('gentypef_fmin','gentypef_fmin',[fmin_gentypef_input_type1,fmin_gentypef_input_type2],fmin_gentypef_output_type,[fmin_gentypef_input_values1,fmin_gentypef_input_values2],'0 * FLT_ULP')
#  
  ##### gentype fmod(gentype x, gentype y)
  fmod_base_values = base_input_values
  fmod_input_values1 = []
  fmod_input_values2 = []
  fmod_input_values1,fmod_input_values2=gene2ValuesLoop(fmod_input_values1,fmod_input_values2,fmod_base_values)
  fmod_input_type1 = ['float','float2','float4','float8','float16']
  fmod_input_type2 = ['float','float2','float4','float8','float16']
  fmod_output_type = ['float','float2','float4','float8','float16']
  fmodUtests = func('fmod','fmod',[fmod_input_type1,fmod_input_type2],fmod_output_type,[fmod_input_values1,fmod_input_values2],'0 * FLT_ULP')
  
  ##### gentype hypot(gentype x, gentype y)
  hypot_base_values = base_input_values
  hypot_input_values1 = []
  hypot_input_values2 = []
  hypot_input_values1,hypot_input_values2=gene2ValuesLoop(hypot_input_values1,hypot_input_values2,hypot_base_values)
  hypot_input_type1 = ['float','float2','float4','float8','float16']
  hypot_input_type2 = ['float','float2','float4','float8','float16']
  hypot_output_type = ['float','float2','float4','float8','float16']
  hypotUtests = func('hypot','hypot',[hypot_input_type1,hypot_input_type2],hypot_output_type,[hypot_input_values1,hypot_input_values2],'4 * FLT_ULP')
  
  ##### intn ilogb(floartn x)
  ilogb_input_values = base_input_values
  ilogb_input_type = ['float','float2','float4','float8','float16']
  ilogb_output_type = ['int','int2','int4','int8','int16']
  ilogbUtests = func('ilogb','ilogb',[ilogb_input_type],ilogb_output_type,[ilogb_input_values],'0 * INT_ULP')

  #### floatn ldexp(floatnx, intnk)
  ldexp_input_values1 = [FLT_MAX_POSI,FLT_MIN_NEGA,FLT_MIN_POSI,FLT_MAX_NEGA,80, -80, 3.14, -3.14, 0.5, 1, 0.0,1500.24,-1500.24]
  ldexp_input_values2 = [-1,-2,-3,4,5,6,7,8,10,12,14,16,12]
  ldexp_input_type1 = ['float','float2','float4','float8','float16']
  ldexp_input_type2 = ['int','int2','int4','int8','int16']
  ldexp_output_type = ['float','float2','float4','float8','float16']
  ldexp_cpu_func='''
namespace utest {
static float ldexp(float x, int y){
    return x * exp2(y);
}
} '''
  ldexpUtests = func('ldexp','utest::ldexp',[ldexp_input_type1,ldexp_input_type2],ldexp_output_type,[ldexp_input_values1,ldexp_input_values2],'0 * FLT_ULP', ldexp_cpu_func)

  ##### gentype lgamma(gentype x)
  lgamma_input_values = base_input_values
  lgamma_input_type = ['float','float2','float4','float8','float16']
  lgamma_output_type = ['float','float2','float4','float8','float16']
  lgammaUtests = func('lgamma','lgamma',[lgamma_input_type],lgamma_output_type,[lgamma_input_values],'16 * FLT_ULP')

  ##### gentype log(gentype)
  log_input_values = base_input_values
  log_input_type = ['float','float2','float4','float8','float16']
  log_output_type = ['float','float2','float4','float8','float16']
  logUtests = func('log','log',[log_input_type],log_output_type,[log_input_values],'4 * FLT_ULP')
  
  ##### gentype log2(gentype)
  log2_input_values = base_input_values
  log2_input_type = ['float','float2','float4','float8','float16']
  log2_output_type = ['float','float2','float4','float8','float16']
  log2Utests = func('log2','log2',[log2_input_type],log2_output_type,[log2_input_values],'4 * FLT_ULP')
  
  ##### gentype log10(gentype)
  log10_input_values = base_input_values
  log10_input_type = ['float','float2','float4','float8','float16']
  log10_output_type = ['float','float2','float4','float8','float16']
  log10Utests = func('log10','log10',[log10_input_type],log10_output_type,[log10_input_values],'4 * FLT_ULP')
  
  ##### gentype log1p(gentype x)
  log1p_input_values = base_input_values
  log1p_input_type = ['float','float2','float4','float8','float16']
  log1p_output_type = ['float','float2','float4','float8','float16']
  log1pUtests = func('log1p','log1p',[log1p_input_type],log1p_output_type,[log1p_input_values],'4 * FLT_ULP')
  
  ##### gentype logb(gentype x)
  logb_input_values = base_input_values
  logb_input_type = ['float','float2','float4','float8','float16']
  logb_output_type = ['float','float2','float4','float8','float16']
  logbUtests = func('logb','logb',[logb_input_type],logb_output_type,[logb_input_values],'0 * FLT_ULP')
  
  ##### gentype maxmag(gentype x, gentype y)
  maxmag_base_values = base_input_values
  maxmag_input_values1 = []
  maxmag_input_values2 = []
  maxmag_input_values1,maxmag_input_values2=gene2ValuesLoop(maxmag_input_values1,maxmag_input_values2,maxmag_base_values)
  maxmag_input_type1 = ['float','float2','float4','float8','float16']
  maxmag_input_type2 = ['float','float2','float4','float8','float16']
  maxmag_output_type = ['float','float2','float4','float8','float16']
  maxmag_cpu_func='''
static float maxmag(float x, float y){
  if(fabs(x) > fabs(y))
    return x;
  else if (fabs(x) < fabs(y))
    return y;
  else
    return fmax(x,y);
} '''
  maxmagUtests = func('maxmag','maxmag',[maxmag_input_type1,maxmag_input_type2],maxmag_output_type,[maxmag_input_values1,maxmag_input_values2],'0 * FLT_ULP',maxmag_cpu_func)
  
  ##### gentype minmag(gentype x, gentype y)
  minmag_base_values = base_input_values
  minmag_input_values1 = []
  minmag_input_values2 = []
  minmag_input_values1,minmag_input_values2=gene2ValuesLoop(minmag_input_values1,minmag_input_values2,minmag_base_values)
  minmag_input_type1 = ['float','float2','float4','float8','float16']
  minmag_input_type2 = ['float','float2','float4','float8','float16']
  minmag_output_type = ['float','float2','float4','float8','float16']
  minmag_cpu_func='''
static float minmag(float x, float y){
  if(fabs(x) < fabs(y))
    return x;
  else if (fabs(x) > fabs(y))
    return y;
  else
    return fmin(x,y);
} '''
  minmagUtests = func('minmag','minmag',[minmag_input_type1,minmag_input_type2],minmag_output_type,[minmag_input_values1,minmag_input_values2],'0 * FLT_ULP',minmag_cpu_func)
  
#  ##### floatn nan(uintn nancode)
#  nan_input_values = base_input_values
#  nan_input_type = ['uint','uint2','uint4','uint8','uint16']
#  nan_output_type = ['float','float2','float4','float8','float16']
#  nanUtests = func('nan','nan',[nan_input_type],nan_output_type,[nan_input_values],'0 * FLT_ULP')
  
  ##### gentype nextafter(gentype x, gentype y)
  nextafter_base_values = base_input_values
  nextafter_input_values1 = []
  nextafter_input_values2 = []
  nextafter_input_values1,nextafter_input_values2=gene2ValuesLoop(nextafter_input_values1,nextafter_input_values2,nextafter_base_values)
  nextafter_input_type1 = ['float','float2','float4','float8','float16']
  nextafter_input_type2 = ['float','float2','float4','float8','float16']
  nextafter_output_type = ['float','float2','float4','float8','float16']
  nextafterUtests = func('nextafter','nextafterf',[nextafter_input_type1,nextafter_input_type2],nextafter_output_type,[nextafter_input_values1,nextafter_input_values2],'0 * FLT_ULP')
  
  ##### gentype pow(gentype x, gentype y)
  pow_base_values = [80, -80, 3.14, -3.14, 0.5, 1, -3,-4,2,0.0,-0.0,1500.24,-1500.24]
  pow_input_values1 = []
  pow_input_values2 = []
  pow_input_values1,pow_input_values2=gene2ValuesLoop(pow_input_values1,pow_input_values2,pow_base_values)
  pow_input_type1 = ['float','float2','float4','float8','float16']
  pow_input_type2 = ['float','float2','float4','float8','float16']
  pow_output_type = ['float','float2','float4','float8','float16']
  pow_cpu_func='''
static float pow_utest(float x, float y){
    if ((x == 0.0f) && (y == -INFINITY))
        return INFINITY;
    return pow(x,y);
} '''
  powUtests = func('pow','pow_utest',[pow_input_type1,pow_input_type2],pow_output_type,[pow_input_values1,pow_input_values2],'16 * FLT_ULP',pow_cpu_func)
  
  ##### floatn pown(floatn x, intn y)
  pown_input_values1 = [FLT_MAX_POSI,FLT_MIN_NEGA,FLT_MIN_POSI,FLT_MAX_NEGA,80, -80, 3.14, -3.14, 0.5, 1, 0.0,1500.24,-1500.24]
  pown_input_values2 = [-1,-2,-3,4,5,6,7,8,10,12,14,16,12]
  pown_input_type1 = ['float','float2','float4','float8','float16']
  pown_input_type2 = ['int','int2','int4','int8','int16']
  pown_output_type = ['float','float2','float4','float8','float16']
  pown_cpu_func='''
static float pown(float x, int y){
    return pow(x,y);
} '''
  pownUtests = func('pown','pown',[pown_input_type1,pown_input_type2],pown_output_type,[pown_input_values1,pown_input_values2],'16 * FLT_ULP', pown_cpu_func)
  
  ##### gentype powr(gentype x, gentype y)
  powr_input_values1 = [80, 80, 3.14, 1, 1.257, +0.0, -0.0, +0.0, -0.0, +0.0, -0.0, +1, +1, 80, +0.0, -0.0, +0.0, -0.0, 'INFINITY','INFINITY', +1, +1, +0.0, 2.5,' NAN', 'NAN', 'NAN']
  powr_input_values2 = [5.5, 6,7, +0.0, -0.0, -1, -15.67, '-INFINITY', '-INFINITY', 1,  -2.7, 10.5, 3.1415, 3.5, -0.0, -0.0, +0.0, +0.0, +0.0, -0.0, 'INFINITY', '-INFINITY', 'NAN', 'NAN', -1.5, +0.0, 1.5]
  powr_input_type1 = ['float','float2','float4','float8','float16']
  powr_input_type2 = ['float','float2','float4','float8','float16']
  powr_output_type = ['float','float2','float4','float8','float16']
  powr_cpu_func='''
static float powr(float x, float y){
    if (((x > 0) && (x != +INFINITY)) && (y == 0.0f))
        return 1;
    else if ((x == 0.0f) && ((y < 0 ) || (y == -INFINITY)))
        return +INFINITY;
    else if ((x == 0.0f) && (y > 0))
        return +0;
    else if ((x == 0.0f) && (y == 0.0f))
        return NAN;
    else if ((x == +1) && ((y == +INFINITY) || (y == -INFINITY)))
        return NAN;
    else if ((x == +1) && ((y != +INFINITY) && (y != -INFINITY)))
        return 1;
    else if ((x == +INFINITY) && (y == 0.0f))
        return NAN;
    else if (std::isnan(x) || (x < 0))
        return NAN;
    else if ((x >=  0) && (std::isnan(y)))
        return NAN;
    else
        return powf(x,y);
} '''
  powrUtests = func('powr','powr',[powr_input_type1,powr_input_type2],powr_output_type,[powr_input_values1,powr_input_values2],'16 * FLT_ULP', powr_cpu_func)
  
  ##### gentype remainder(gentype x, gentype y)
  remainder_base_values = base_input_values
  remainder_input_values1 = []
  remainder_input_values2 = []
  remainder_input_values1,remainder_input_values2=gene2ValuesLoop(remainder_input_values1,remainder_input_values2,remainder_base_values)
  remainder_input_type1 = ['float','float2','float4','float8','float16']
  remainder_input_type2 = ['float','float2','float4','float8','float16']
  remainder_output_type = ['float','float2','float4','float8','float16']
  remainderUtests = func('remainder','remainder',[remainder_input_type1,remainder_input_type2],remainder_output_type,[remainder_input_values1,remainder_input_values2],'0 * FLT_ULP')
  
  ##### gentype rint(gentype x)
  rint_input_values = base_input_values
  rint_input_type = ['float','float2','float4','float8','float16']
  rint_output_type = ['float','float2','float4','float8','float16']
  rintUtests = func('rint','rint',[rint_input_type],rint_output_type,[rint_input_values],'0 * FLT_ULP')
  
  ##### floatn rootn(floatn x, intn y)
  rootn_input_values1 = [0.0, 0.0012,  0.5, 1, 3.14, 12345]
  rootn_input_values2 = [-1, 1, -20, 20, -123, 456]
  rootn_input_type1 = ['float','float2','float4','float8','float16']
  rootn_input_type2 = ['int','int2','int4','int8','int16']
  rootn_output_type = ['float','float2','float4','float8','float16']
  rootn_cpu_func='''
static float rootn(float x, int y){
    return pow(x,1.0/y);
} '''
  rootnUtests = func('rootn','rootn',[rootn_input_type1,rootn_input_type2],rootn_output_type,[rootn_input_values1,rootn_input_values2],'4 * FLT_ULP',rootn_cpu_func)
  
  ##### gentype round(gentype x)
  round_input_values = base_input_values
  round_input_type = ['float','float2','float4','float8','float16']
  round_output_type = ['float','float2','float4','float8','float16']
  roundUtests = func('round','round',[round_input_type],round_output_type,[round_input_values],'0 * FLT_ULP')
  
  ##### gentype rsqrt(gentype)
  rsqrt_input_values = base_input_values
  rsqrt_input_type = ['float','float2','float4','float8','float16']
  rsqrt_output_type = ['float','float2','float4','float8','float16']
  rsqrt_cpu_func='''
static float rsqrt(float x)
{ return 1/sqrt(x);} '''
  rsqrtUtests = func('rsqrt','rsqrt',[rsqrt_input_type],rsqrt_output_type,[rsqrt_input_values],'4 * FLT_ULP', rsqrt_cpu_func)

 
  ##### gentype sin(gentype)
  sin_input_values = base_input_values
  sin_input_type = ['float','float2','float4','float8','float16']
  sin_output_type = ['float','float2','float4','float8','float16']
  sinUtests = func('sin','sin',[sin_input_type],sin_output_type,[sin_input_values],'4 * FLT_ULP')
  
#  ##### gentype sincos(gentype)
#  sincos_input_values1 = [FLT_MAX_POSI,FLT_MIN_NEGA,FLT_MIN_POSI,FLT_MAX_NEGA,80, -80, 3.14, -3.14, -0.5, 0.5, 1, -1, 0.0,6,-6,1500.24,-1500.24]
#  sincos_input_values2 = []
#  sincos_input_type1 = ['float','float2','float4','float8','float16']
#  sincos_input_type2 = ['float','float2','float4','float8','float16']
#  sincos_output_type = ['float','float2','float4','float8','float16']
#  ###### gentype sincos(gentype)
#  #  sincosUtests = func('sincos','sincos',[sincos_input_type1,sincos_input_type2],sincos_output_type,[sincos_input_values1,sincos_input_values2],'4 * FLT_ULP')
  
  ##### gentype sinh(gentype)
  sinh_input_values = base_input_values
  sinh_input_type = ['float','float2','float4','float8','float16']
  sinh_output_type = ['float','float2','float4','float8','float16']
  sinhUtests = func('sinh','sinh',[sinh_input_type],sinh_output_type,[sinh_input_values],'4 * FLT_ULP')
  
  ##### gentype sinpi(gentype x)
  sinpi_input_values = [0, 1, 3.14, -0.88, -0.12, -0.5, 0.5, -0.49, 0.49, 0.51, -0.51, -0.1, 0.1]
  sinpi_input_type = ['float','float2','float4','float8','float16']
  sinpi_output_type = ['float','float2','float4','float8','float16']
  sinpi_cpu_func=reduce1+sinpi
  sinpiUtests = func('sinpi','sinpi',[sinpi_input_type],sinpi_output_type,[sinpi_input_values],'4 * FLT_ULP',sinpi_cpu_func)
  
  ##### gentype sqrt(gentype)
  sqrt_input_values = base_input_values
  sqrt_input_type = ['float','float2','float4','float8','float16']
  sqrt_output_type = ['float','float2','float4','float8','float16']
  sqrtUtests = func('sqrt','sqrt',[sqrt_input_type],sqrt_output_type,[sqrt_input_values],'4 * FLT_ULP')
  
  ##### gentype tan(gentype)
  tan_input_values = base_input_values
  tan_input_type = ['float','float2','float4','float8','float16']
  tan_output_type = ['float','float2','float4','float8','float16']
  tanUtests = func('tan','tan',[tan_input_type],tan_output_type,[tan_input_values],'5 * FLT_ULP')
  
  ##### gentype tanh(gentype)
  tanh_input_values = base_input_values
  tanh_input_type = ['float','float2','float4','float8','float16']
  tanh_output_type = ['float','float2','float4','float8','float16']
  tanhUtests = func('tanh','tanh',[tanh_input_type],tanh_output_type,[tanh_input_values],'5 * FLT_ULP')
  
  ##### gentype tanpi(gentype x)
  tanpi_input_values = [ 0, 3.14, 5.15, 6.01, 7.89]
  tanpi_input_type = ['float','float2','float4','float8','float16']
  tanpi_output_type = ['float','float2','float4','float8','float16']
  tanpi_cpu_func=reduce1+sinpi+cospi+'''
static float tanpi(float x){
  return sinpi(x)/cospi(x);
}
'''
  tanpiUtests = func('tanpi','tanpi',[tanpi_input_type],tanpi_output_type,[tanpi_input_values],'400 * FLT_ULP',tanpi_cpu_func)
  
  ##### gentype trunc(gentype)
  trunc_input_values = base_input_values
  trunc_input_type = ['float','float2','float4','float8','float16']
  trunc_output_type = ['float','float2','float4','float8','float16']
  truncUtests = func('trunc','trunc',[trunc_input_type],trunc_output_type,[trunc_input_values],'0 * FLT_ULP')

if __name__ == "__main__":
  main()
